open Prolog

let search program goal sk fk =
  let rec search goal sk fk = match goal with
    | Atom (p,l) -> search (lookup program p l) sk fk
    | Or (a,b) -> search a sk (fun () -> search b sk fk)
    | And (a,b) -> search a (fun k -> search b sk k) fk
    | Equals (u,v) ->
        let s = Term.save () in
        let unifies =
          try Unify.unify u v ; true with
            | Unify.Unification_failure -> false
        in
          if unifies then
            sk (fun () -> Term.restore s ; fk ())
          else begin
            Term.restore s ;
            fk ()
          end
    | False -> fk ()
    | True -> sk fk
  in
    search goal sk fk

let has_solution program goal =
  let s = Term.save () in
    search program goal (fun _ -> Term.restore s ; true) (fun () -> false)

let search_interactive program goal =
  let s = Term.save () in
    search program goal
      (fun k ->
         Printf.printf "Solution Found. More? %!" ;
         let l = read_line () in
           if l = "" || l.[0] = 'y' || l.[0] = 'Y' then
             k ()
           else
             Term.restore s)
      (fun () ->
         Printf.printf "No (more) solution.\n")
