open Prolog

type sk = goal list

type fk = (Term.state * sk) list

let cut_mark = Term.save (), [Atom("?",[])]

let rec cut n = function
  | (_,[Atom("?",[])])::_ as fk -> (* Format.printf "Cut %d@." n ; *) fk
  | hd::tl -> cut (n+1) tl
  | [] -> assert false

let cut fk = cut 0 fk

let rec uncut = function
  | (_,[Atom("?",[])])::fk -> uncut fk
  | fk -> fk

let search program sk fk =

  let rec search goals fk = match goals with

    | Atom ("cut",[]) :: goals -> search goals (cut fk)

    | Atom (p,l) :: goals ->
        search (lookup program p l :: goals) (cut_mark::fk)

    | Or (a,b) :: goals ->
        search (a :: goals) ((Term.save (), b :: goals) :: fk)
    | And (a,b) :: goals ->
        search (a :: b :: goals) fk
    | Equals (u,v) :: goals ->
        let unifies =
          try Unify.unify u v ; true with Unify.Unification_failure -> false
        in
          if unifies then
            search goals fk
          else begin
            match uncut fk with
              | [] -> false
              | (s,sk)::fk -> Term.restore s ; search sk fk
          end
    | False :: _ ->
        begin match uncut fk with
          | [] -> false
          | (s,sk)::fk -> Term.restore s ; search sk fk
        end
    | True :: goals -> search goals fk
    | [] -> true
  in
    search sk fk

let has_solution program goal =
  let s = Term.save () in
  let r = search program [goal] [s,[False]] in
    if r then Term.restore s ;
    r
