open Prolog

type sk = Conj of goal * sk | Success
type fk = Disj of goal * sk * fk | Restore of Term.state * fk | Failure

let search program goal sk fk =

  let rec eval_sk sk fk = match sk with
    | Success -> true
    | Conj (g,sk) -> search g sk fk

  and eval_fk = function
    | Failure -> false
    | Disj (g,sk,fk) -> search g sk fk
    | Restore (s,fk) -> Term.restore s ; eval_fk fk

  and search goal (sk:sk) (fk:fk) = match goal with
    | Atom (p,l) -> search (lookup program p l) sk fk
    | Or (a,b) -> search a sk (Disj (b,sk,fk))
    | And (a,b) -> search a (Conj (b,sk)) fk
    | Equals (u,v) ->
        let s = Term.save () in
        let unifies =
          try Unify.unify u v ; true with
            | Unify.Unification_failure -> false
        in
          if unifies then
            eval_sk sk (Restore (s,fk))
          else begin
            Term.restore s ;
            eval_fk fk
          end
    | False -> eval_fk fk
    | True -> eval_sk sk fk
  in
    search goal sk fk

let has_solution program goal =
  let s = Term.save () in
  let r = search program goal Success Failure in
    if r then Term.restore s ;
    r
