
type args = Term.in_term list

type goal =
  | Atom of string * args
  | Equals of Term.in_term * Term.in_term
  | And of goal * goal
  | Or of goal * goal
  | False
  | True

type predicate = args -> goal
module SMap = Map.Make(String)
type program = predicate SMap.t

let lookup prog p args =
  try (SMap.find p prog) args with
    | Not_found -> False

let prog_empty =
  SMap.singleton "print"
    (fun args ->
       assert (List.length args = 1) ;
       Format.printf "%a@." Term.pp (List.hd args) ;
       True)

let concat_prog p1 p2 =
  SMap.merge
    (fun p a b -> match a,b with
       | Some x, None | None, Some x -> Some x
       | _ -> failwith ("multiple définition de " ^ p))
    p1 p2

type clause = args -> goal

let make_predicate p clauses =
  SMap.singleton p
    (fun l ->
       List.fold_left
         (fun goal clause -> Or (goal, clause l))
         False
         clauses)

let forall f =
  fun args ->
    let v = Term.fresh () in
      f (Term.var v) args

let horn hd body : clause =
  fun args ->
    match hd with
      | Atom (_,l) ->
          List.fold_left2
            (fun goal a x -> And (Equals (a,x), goal))
            body
            args l
      | _ -> assert false
